from abc import ABC
from abc import abstractmethod


class Step(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def process(self, data, inputs, utils):
        pass


class StepException(Exception):
    pass



import sys
import getopt
import logging


from yt_concate.pipeline.steps.preflight import Preflight
from yt_concate.pipeline.steps.get_video_list import GetVideoList
from yt_concate.pipeline.steps.initialize_yt import InitializeYT
from yt_concate.pipeline.steps.download_caption import DownLoadCaptions
from yt_concate.pipeline.steps.read_caption import ReadCaption
from yt_concate.pipeline.steps.search import Search
from yt_concate.pipeline.steps.download_videos import DownloadVideos
from yt_concate.pipeline.steps.edit_video import EditVideo
from yt_concate.pipeline.steps.postflight import Postflight
from yt_concate.pipeline.steps.step import StepException
from yt_concate.pipeline.pipeline import Pipeline
from yt_concate.utils import Utils


def print_usage():
    print("python main.py -c <channel_id> -s <search_word> -l <limit>")
    print("python main.py --channel_id <channel_id> --search_word <search_word> --limit <limit>")
    print("options")
    print("{:>5}{:<12} ,{}".format("-c", "--channel_id", "channel id for youtube channel"))
    print("{:>5}{:<12} ,{}".format("-s", "--search_word", "search world in the channel videos"))
    print("{:>5}{:<12} ,{}".format("-l", "--limit", "integer,quantity for concatenating videos"))
    print("{:>5}{:<12} ,{}".format("", "--cleanup", "logical,delete files after the result files complete"))
    print("{:>5}{:<12} ,{}".format("", "--level", "the level to print on the screen,default is logging.INFO"))


# channel_id = "UCKSVUHI9rbbkXhvAXK-2uxA"
# search_word="incredible"


def main():
    inputs = {
        "channel_id": "",
        "search_cord": "",
        "limit": "",
        "cleanup": True,
        "level": logging.INFO,
    }
    short_opt = "hc:s:l:"
    long_opt = "help channel_id= search_word= limit= cleanup= level=".split()
    try:
        opts, args = getopt.getopt(sys.argv[1:], short_opt, long_opt)
    except getopt.GetoptError:
        print_usage()
        sys.exit(2)

    for opt, arg in opts:
        if opt in ('-h', "help"):
            print_usage()
            sys.exit(0)
        elif opt in ("-c", "--channel_id"):
            inputs["channel_id"] = arg
        elif opt in ("-s", "--search_cord"):
            inputs["search_cord"] = arg
        elif opt in ("-l", "--limit"):
            inputs["limit"] = arg
        elif opt == "cleanup":
            inputs["cleanup"] = arg
        elif opt == "level":
            inputs["level"] = arg
    if not inputs["limit"].isnumeric():
        print_usage()
        sys.exit(2)
    if not inputs["cleanup"] in ("True", "true", "False", "false"):
        print_usage()
        sys.exit(2)

    steps = [
        Preflight(),
        GetVideoList(),
        InitializeYT(),
        DownLoadCaptions(),
        ReadCaption(),
        Search(),
        DownloadVideos(),
        EditVideo(),
        Postflight(),
    ]
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    file_handler = logging.FileHandler("project.log")
    formatter = logging.Formatter("%(levelname)s:%(asctime)s:%(message)s")
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)

    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(inputs["level"])
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)

    utils = Utils()
    p = Pipeline(steps)
    p.run(inputs, utils)


if __name__ == "__main__":
    main()



import logging

from .steps.step import StepException


class Pipeline:
    def __init__(self, steps):
        self.steps = steps

    def run(self, inputs, utils):
        data = None
        for step in self.steps:
            try:
                data = step.process(data, inputs, utils)
            except StepException as e:
                logging.warning("Exception happened:", e)
                break



import logging

from yt_concate.pipeline.steps.step import Step


class Preflight(Step):
    def process(self, data, inputs, utils):
        logger = logging.getLogger()
        logger.info("in preflight")
        utils.create_dir()



import os

from dotenv import load_dotenv

load_dotenv()
API_KEY = os.getenv("api_key")

DOWNLOADS_DIR = "downloads"
VIDEOS_DIR = os.path.join(DOWNLOADS_DIR, "videos")  # downloads/videos
CAPTIONS_DIR = os.path.join(DOWNLOADS_DIR, "captions")  # downloads/captions
OUTPUTS_DIR = "outputs"



import ast

from yt_concate.pipeline.steps.step import Step


class ReadCaption(Step):
    def process(self, data, inputs, utils):
        for yt in data:
            if not utils.caption_file_exist(yt):
                continue

            with open(yt.caption_filepath, "r") as f:
                captions = {}
                data_file = ast.literal_eval(f.read())
                for line in data_file:
                    caption = line['text']
                    time = str(line["start"]) + "-->" + str(line["duration"])
                    captions[caption] = time
            yt.captions = captions
        return data



from yt_concate.pipeline.steps.step import Step
from yt_concate.model.found import Found


class Search(Step):
    def process(self, data, inputs, utils):
        search_word = inputs["search_word"]
        found = []
        for yt in data:
            captions = yt.captions
            if not captions:
                continue
            for caption in captions:
                if search_word in caption:
                    time = captions[caption]
                    f = Found(yt, caption, time)
                    found.append(f)
        return found



import os

from yt_concate.setting import CAPTIONS_DIR
from yt_concate.setting import DOWNLOADS_DIR
from yt_concate.setting import VIDEOS_DIR
from yt_concate.setting import OUTPUTS_DIR


class Utils:
    def __init__(self):
        pass

    def create_dir(self):
        os.makedirs(DOWNLOADS_DIR, exist_ok=True)
        os.makedirs(CAPTIONS_DIR, exist_ok=True)
        os.makedirs(VIDEOS_DIR, exist_ok=True)
        os.makedirs(OUTPUTS_DIR, exist_ok=True)

    def caption_file_exist(self, yt):
        path = yt.caption_filepath
        return os.path.exists(path) and os.path.getsize(path) > 0

    def video_file_exist(self, yt):
        path = yt.video_filepath
        return os.path.exists(path) and os.path.getsize(path) > 0

    def get_video_list_filepath(self, channel_id):
        return os.path.join(DOWNLOADS_DIR, channel_id)

    def video_list_file_exist(self, channel_id):
        path = self.get_video_list_filepath(channel_id)
        return os.path.exists(path) and os.path.getsize(path) > 0

    def output_filepath(self, channel_id, search_word):
        filename = channel_id + "_" + search_word + ".mp4"
        return os.path.join(OUTPUTS_DIR, filename)



import os
import logging

from yt_concate.pipeline.steps.step import Step
from yt_concate.setting import VIDEOS_DIR, CAPTIONS_DIR


class Postflight(Step):
    def process(self, data, inputs, utils):
        logger = logging.getLogger()
        logger.info("in postflight")
        if inputs["cleanup"] == "True":
            os.remove(VIDEOS_DIR)
            os.remove(CAPTIONS_DIR)



import urllib.request
import json
import logging

from yt_concate.setting import API_KEY
from yt_concate.pipeline.steps.step import Step


class GetVideoList(Step):
    def process(self, data, inputs, utils):
        logger = logging.getLogger()
        channel_id = inputs["channel_id"]
        if utils.video_list_file_exist(channel_id):
            logger.info('found existing video list file for channel id', channel_id)
            return self.read_file(utils.get_video_list_filepath(channel_id))
        base_video_url = 'https://www.youtube.com/watch?v='
        base_search_url = 'https://www.googleapis.com/youtube/v3/search?'

        first_url = base_search_url + 'key={}&channelId={}&part=snippet,id&order=date&maxResults=25'.format(API_KEY,
                                                                                                            channel_id)

        video_links = []
        url = first_url
        while True:
            inp = urllib.request.urlopen(url)
            resp = json.load(inp)

            for i in resp['items']:
                if i['id']['kind'] == "youtube#video":
                    video_links.append(base_video_url + i['id']['videoId'])

            try:
                next_page_token = resp['nextPageToken']
                url = first_url + '&pageToken={}'.format(next_page_token)
            except KeyError:
                continue
        logger.info(video_links)
        self.write_to_file(video_links, utils.get_video_list_filepath(channel_id))
        return video_links

    def write_to_file(self, video_links, filepath):
        with open(filepath, "w") as f:
            for url in video_links:
                f.write(url + "\n")

    def read_file(self, filepath):
        video_links = []
        with open(filepath, "r") as f:
            for url in f:
                video_links.append(url.strip())
        return video_links



import os
import logging

from pytube import YouTube
from threading import Thread
from yt_concate.pipeline.steps.step import Step
from yt_concate.setting import VIDEOS_DIR


class DownloadVideos(Step):
    def process(self, data, inputs, utils):
        logger = logging.getLogger()
        yt_set = set([found.yt for found in data])
        logger.info('video to download', len(yt_set))
        for yt in yt_set:
            url = yt.url
            if utils.video_file_exist(yt):
                logger.info(f"found existing video file:{url},skipping")
                continue
            logger.info("downloading", url)

            self.process_thread(data)
            threads = []
            for i in range(os.cpu_count()):
                threads.append(Thread(target=self.process_thread))
            for thread in threads:
                thread.start()
            for thread in threads:
                thread.join()
        return data

    def process_thread(self, data):
        for yt in data:
            YouTube(yt.url).streams.first().download(output_path=VIDEOS_DIR, filename=yt.id + ".mp4")



import os
import time
import logging

from threading import Thread
from youtube_transcript_api import YouTubeTranscriptApi
from yt_concate.pipeline.steps.step import Step


class DownLoadCaptions(Step):
    def process(self, data, inputs, utils):
        logger = logging.getLogger()
        start = time.time()
        for yt in data:
            logger.info("downloading caption for url", yt.id)
            if utils.caption_file_exist(yt):
                logger.info("found existing file")
                continue
            try:
                self.process_thread(data)
                threads = []
                for i in range(os.cpu_count()):
                    threads.append(Thread(target=self.process_thread))
                for thread in threads:
                    thread.start()
                for thread in threads:
                    thread.join()
                with open(yt.caption_filepath, "w") as f:
                    for i in str(threads):
                        f.write(i)

            except Exception:
                logger.warning('error when downloading caption for', yt.url)
                continue
        end = time.time()
        logger.info("took", end - start, "seconds")
        return data

    def process_thread(self, data):
        for yt in data:
            YouTubeTranscriptApi.get_transcript(yt.url)



import math

from moviepy.editor import VideoFileClip, concatenate_videoclips
from yt_concate.pipeline.steps.step import Step


class EditVideo(Step):
    def process(self, data, inputs, utils):
        clips = []
        for found in data:
            start, end = self.parse_time(found.time)
            video = VideoFileClip(found.yt.video_filepath).subclip(start, end)
            clips.append(video)
            if len(clips) >= inputs["limit"]:
                break
        final_clip = concatenate_videoclips(clips)
        output_file = utils.output_filepath(inputs["channel_id"], inputs["search_word"])
        final_clip.write_videofile(output_file, audio_codec='aac')

    def parse_time(self, captions_time):
        a, b = captions_time.split("-->")
        start = eval(a)
        end = start + eval(b)
        return self.calculate_time(start), self.calculate_time(end)

    def calculate_time(self, time):
        if time >= 3600:
            hour = math.floor(time / 3600)
            minute = math.floor((time - 3600 * hour) / 60)
            second = (time - 3600 * hour) % 60
            return hour, minute, second  # return tuple
        else:
            minute = math.floor(time / 60)
            second = time % 60
            return 0, minute, second  # return tuple
